@@ -12,12 +12,18 @@
 define([
     'jquery',
     'base/js/utils',
+    'base/js/i18n',
     'codemirror/lib/codemirror',
     'codemirror/addon/edit/matchbrackets',
     'codemirror/addon/edit/closebrackets',
-    'codemirror/addon/comment/comment'
-], function($, utils, CodeMirror, cm_match, cm_closeb, cm_comment) {
+    'codemirror/addon/comment/comment',
+    'services/config',
+], function($, utils, i18n, CodeMirror, cm_match, cm_closeb, cm_comment, configmod) {
     "use strict";
+
+    function is_single_cursor(dict1, dict2) {
+        return ((dict1.line == dict2.line) && (dict1.ch == dict2.ch));
+    };
     
     var overlayHack = CodeMirror.scrollbarModel.native.prototype.overlayHack;
 
@@ -49,10 +55,9 @@ define([
         options = options || {};
         this.keyboard_manager = options.keyboard_manager;
         this.events = options.events;
-        var config = utils.mergeopt(Cell, options.config);
+        var config = options.config;
         // superclass default overwrite our default
         
-        this.placeholder = config.placeholder || '';
         this.selected = false;
         this.anchor = false;
         this.rendered = false;
@@ -71,10 +76,11 @@ define([
             }
         });
 
+
         // backward compat.
         Object.defineProperty(this, 'cm_config', {
             get: function() {
-                console.warn("Warning: accessing Cell.cm_config directly is deprecated.");
+                console.warn(i18n.msg._("Warning: accessing Cell.cm_config directly is deprecated."));
                 return that._options.cm_config;
             },
         });
@@ -82,14 +88,21 @@ define([
         // load this from metadata later ?
         this.user_highlight = 'auto';
 
-
-        var _local_cm_config = {};
-        if(this.class_config){
-            _local_cm_config = this.class_config.get_sync('cm_config');
+        // merge my class-specific config data with general cell-level config
+        var class_config_data = {};
+        if (this.class_config) {
+            class_config_data = this.class_config.get_sync();
         }
-        config.cm_config = utils.mergeopt({}, config.cm_config, _local_cm_config);
+
+        var cell_config = new configmod.ConfigWithDefaults(options.config,
+            Cell.options_default, 'Cell');
+        var cell_config_data = cell_config.get_sync();
+
+        // this._options is a merge of SomeCell and Cell config data:
+        this._options = utils.mergeopt({}, cell_config_data, class_config_data);
+        this.placeholder = this._options.placeholder || '';
+
         this.cell_id = utils.uuid();
-        this._options = config;
 
         // For JS VM engines optimization, attributes should be all set (even
         // to null) in the constructor, and if possible, if different subclass
@@ -100,6 +113,12 @@ define([
         this.cell_type = this.cell_type || null;
         this.code_mirror = null;
 
+        // The nbformat only specifies attachments for textcell, but to avoid
+        // data loss when switching between cell types in the UI, all cells
+        // have an attachments property here. It is only saved to disk
+        // for textcell though (in toJSON)
+        this.attachments = {};
+
         this.create_element();
         if (this.element !== null) {
             this.element.data("cell", this);
@@ -114,9 +133,13 @@ define([
             readOnly: false,
             theme: "default",
             extraKeys: {
-                "Cmd-Right":"goLineRight",
-                "End":"goLineRight",
-                "Cmd-Left":"goLineLeft"
+                "Cmd-Right": "goLineRight",
+                "End": "goLineRight",
+                "Cmd-Left": "goLineLeft",
+                "Tab": "indentMore",
+                "Shift-Tab" : "indentLess",
+                "Cmd-/" : "toggleComment",
+                "Ctrl-/" : "toggleComment",
             }
         }
     };
@@ -183,6 +206,7 @@ define([
         });
         if (this.code_mirror) {
             this.code_mirror.on("change", function(cm, change) {
+                that.events.trigger("change.Cell", {cell: that, change: change});
                 that.events.trigger("set_dirty.Notebook", {value: true});
             });
         }
@@ -267,6 +291,9 @@ define([
             this.element.addClass('selected');
             this.element.removeClass('unselected');
             this.selected = true;
+            // disable 'insert image' menu item (specific cell types will enable
+            // it in their override select())
+            this.notebook.set_insert_image_enabled(false);
             return true;
         } else {
             return false;
@@ -444,14 +471,14 @@ define([
     };
 
     /**
-     * should be overritten by subclass
+     * should be overwritten by subclass
      * @method get_text
      */
     Cell.prototype.get_text = function () {
     };
 
     /**
-     * should be overritten by subclass
+     * should be overwritten by subclass
      * @method set_text
      * @param {string} text
      */
@@ -459,7 +486,7 @@ define([
     };
 
     /**
-     * should be overritten by subclass
+     * should be overwritten by subclass
      * serialise cell to json.
      * @method toJSON
      **/
@@ -467,23 +494,39 @@ define([
         var data = {};
         // deepcopy the metadata so copied cells don't share the same object
         data.metadata = JSON.parse(JSON.stringify(this.metadata));
+        if (this.id !== undefined) {
+            data.id = this.id;
+        }
+        if (data.metadata.deletable) {
+            delete data.metadata.deletable;
+        }
+        if (data.metadata.editable) {
+            delete data.metadata.editable;
+        }
+        if (data.metadata.collapsed === false) {
+            delete data.metadata.collapsed;
+        }
         data.cell_type = this.cell_type;
         return data;
     };
 
     /**
-     * should be overritten by subclass
+     * should be overwritten by subclass
      * @method fromJSON
      **/
     Cell.prototype.fromJSON = function (data) {
         if (data.metadata !== undefined) {
             this.metadata = data.metadata;
         }
+        if (data.id !== undefined) {
+            this.id = data.id;
+        }
     };
 
 
     /**
      * can the cell be split into two cells (false if not deletable)
+     *
      * @method is_splittable
      **/
     Cell.prototype.is_splittable = function () {
@@ -500,14 +543,28 @@ define([
     };
 
     /**
-     * is the cell deletable? only false (undeletable) if
-     * metadata.deletable is explicitly false -- everything else
+     * is the cell edtitable? only false (readonly) if
+     * metadata.editable is explicitly false -- everything else
      * counts as true
      *
+     * @method is_editable
+     **/
+    Cell.prototype.is_editable = function () {
+        if (this.metadata.editable === false) {
+            return false;
+        }
+        return true;
+    };
+
+    /**
+     * is the cell deletable? only false (undeletable) if
+     * metadata.deletable is explicitly false or if the cell is not
+     * editable -- everything else counts as true
+     *
      * @method is_deletable
      **/
     Cell.prototype.is_deletable = function () {
-        if (this.metadata.deletable === false) {
+        if (this.metadata.deletable === false || !this.is_editable()) {
             return false;
         }
         return true;
@@ -539,6 +596,80 @@ define([
         return text;
     };
 
+
+    /**
+     * @return {Array} - the text between cursors and within selections (multicursor/sorted)
+     * @method get_split_text
+     **/
+    Cell.prototype.get_split_text = function () {
+        var start = {line:0, ch:0};
+        var last_line_num = this.code_mirror.lineCount()-1;
+        var last_line_len = this.code_mirror.getLine(last_line_num).length;
+        var end = {line:last_line_num, ch:last_line_len};
+
+        var flag_empty_cell = is_single_cursor(start, end);
+        var flag_first_position = false;
+        var flag_last_position = false;
+        var flag_all_select = false;
+
+        var ranges = this.code_mirror.listSelections();
+
+        var cursors = [start];
+
+        for (var i = 0; i < ranges.length; i++) {
+            // append both to handle selections
+            // ranges[i].head.sticky is null if ctrl-a select
+            if ((ranges[i].head.sticky == 'before') || (ranges[i].head.sticky === null )) {
+                cursors.push(ranges[i].anchor);
+                cursors.push(ranges[i].head);
+                if (is_single_cursor(ranges[i].anchor, start) &&
+                    is_single_cursor(ranges[i].head, end)) {
+                    flag_all_select = true;
+                }
+            } else {
+                cursors.push(ranges[i].head);
+                cursors.push(ranges[i].anchor);
+                if (is_single_cursor(ranges[i].head, start) &&
+                    is_single_cursor(ranges[i].anchor, end)) {
+                    flag_all_select = true;
+                }
+            }
+            // single cursor at beginning or end of cell
+            if (is_single_cursor(ranges[i].head, ranges[i].anchor)) {
+                if (is_single_cursor(ranges[i].head, start)) flag_first_position = true;
+                if (is_single_cursor(ranges[i].head, end)) flag_last_position = true;
+            }
+        }
+        cursors.push(end);
+
+        // Cursors is now sorted, but likely has duplicates due to anchor and head being the same for cursors
+        var locations = [cursors[0]];
+        for (var i = 1; i < cursors.length; i++) {
+            var last = locations[locations.length-1];
+            var current = cursors[i];
+            if ((last.line != current.line) || (last.ch != current.ch)) {
+                locations.push(cursors[i]);
+            }
+        }
+
+        // Split text
+        var text_list = [];
+        // Split single cursors at first position
+        if (flag_empty_cell || flag_first_position) text_list.push('');
+        for (var i = 1; i < locations.length; i++) {
+            var text = this.code_mirror.getRange(locations[i-1], locations[i]);
+            text = text.replace(/^\n+/, '').replace(/\n+$/, ''); // removes newlines at beginning and end
+            text_list.push(text);
+        }
+        // Split single cursors at last position
+        if (flag_last_position) text_list.push('');
+        // Duplicate cell if full cell is selected
+        if ((text_list.length == 1) && flag_all_select && !flag_empty_cell) {
+            text_list = text_list.concat(text_list);
+        }
+        return text_list;
+    };
+
     /**
      * Show/Hide CodeMirror LineNumber
      * @method show_line_numbers
@@ -711,7 +842,7 @@ define([
         } else {
             data.metadata = this.metadata;
         }
-        this.element.find('.inner_cell').find("a").text("Unrecognized cell type: " + data.cell_type);
+        this.element.find('.inner_cell').find("a").text(i18n.msg.sprintf(i18n.msg._("Unrecognized cell type: %s"), data.cell_type));
     };
     
     UnrecognizedCell.prototype.create_element = function () {
@@ -725,12 +856,12 @@ define([
         inner_cell.append(
             $("<a>")
                 .attr("href", "#")
-                .text("Unrecognized cell type")
+                .text(i18n.msg._("Unrecognized cell type"))
         );
         cell.append(inner_cell);
         this.element = cell;
     };
-    
+
     UnrecognizedCell.prototype.bind_events = function () {
         Cell.prototype.bind_events.apply(this, arguments);
         var cell = this;
